/* @flow */
import * as React from 'react';
import PropTypes            from 'prop-types';
import invariant            from 'invariant';

// airflux
import Listener from './Listener';
import type Publisher from './Publisher';
import type Action from './Action';
import type Store from './Store';
import Environment from './Environment';
import type { AirfluxApi } from './AirfluxApi';
import * as _ from './utils';


type StoreState = any;

type ConnectStoreProps = {
  environment?: Environment;

  /** Render function that you have to provide */
  render: ( storeStates: { [ storeName: string ]: StoreState } ) => React.Node;
  
  /** stores passed directly should be an object, key being the name of store state, as it will be passed to the render function */
  stores?: { [key: string ]: Store< any > } | string[];

  /** Future use only. */
  query?: ( stores: { [ name: string ]: any } ) => { [ name: string ]: any };
};



/**
 * Connect store is a React components, which serves as a purpose of connecting to a store, and passing the state of stores
 * as props in a child component.
 * In that way, you seperate the connection logic from the component. You also avoid mixing inside the state of one component
 * data coming from stores.
 * 
 * You have two ways of using this component:
 * - passing the stores directly to ConnectStore
 * - using it in conjonction with AirfluxApp
 * 
 * Recommended way of using it, but that's just a suggestion:
 * Create a React component, with the same name of the file in which it is declared; export default a stateless component
 * with the ConnectStore.
 * 
 * 
 * @example
 *   const sampleStore = new SampleStore();
 *   export default () => <airflux.ConnectStore stores={ { sampleStore } } render={ ( { sampleStore } ) => <YouComponent sampleStoreState={ sampleStore } /> } />;
 * 
 * @example
 *   // file: YouComponent.jsx
 *   // when used with AirfluxApp in your routing hierarchy.
 *   export default () => <airflux.ConnectStore render={ ( { sampleStore } ) => <YouComponent sampleStoreState={ sampleStore } /> } />;
 * 
 *   // file: App.jsx
 *   render() {
 *       return (
 *           <airflux.AirfluxApp environment={ appEnv }>
 *              <HashRouter>
 *                  <Switch>
 *                      <Route path="/" component={ YouComponent } />
 *                  </Switch>
 *              </HashRouter>
 *           </airflux.AirfluxApp>
 *       );
 *   }
 */
export default class ConnectStore extends React.Component< ConnectStoreProps, { [ name: string ]: any }> {
    static contextTypes = { airflux: PropTypes.object };

    /**
     * @private
     */
    _listener: Listener = new Listener();

    constructor( props: ConnectStoreProps, context: any ) {
        super( props );

        var publishers = this._getEnvironment( props, context ).getPublishers();

        if( Object.keys( publishers ).length > 0 ) {
            this.state = Object.keys( publishers )
                .filter( this._filterStores( props ) )
                .map( ( name ) => ( { [name]: publishers[ name ].state } ) )
                .reduce( ( a, b ) => Object.assign( {}, a, b ), {} );
        }
    }

    _filterStores( props: ConnectStoreProps ) : ( storeName: string ) => boolean {
        if( Array.isArray( props.stores ) ) {
            return ( storeName ) => props.stores.includes( storeName );
        }
        else {
            return () => true;
        }
    }

    /**
     * Recreates an environment, based on the options we have:
     * - stores, passed directly as props to this component.
     * - environment, passed as a React Context by an AirfluxApp component, up above in the hierarchy
     * - environmnet, passed directly to this component
     * 
     * @private
     * 
     * @param {*} p 
     * @param {*} c 
     */
    _getEnvironment( p?: ConnectStoreProps, c?: any ): Environment {
        const props = p || this.props;
        const context = c || this.context;

        if( !!props.stores && !Array.isArray( props.stores ) ) {
            return new Environment( props.stores );
        }
        else if( !!context && !!context.airflux ) {
            return context.airflux.environment;
        }
        else if( !!props && !!props.environment ) {
            return props.environment;
        }
        else {
            return new Environment();
        }
    }

    getPublishers() : { [key: string ]: Store< any > } {
        return this._getEnvironment().getPublishers();
    }

    /**
     * Start listening to all stores only once the component is mounted.
     */
    componentDidMount() {
        var publishers = this._getEnvironment().getPublishers();

        Object.keys( publishers )
            .filter( this._filterStores( this.props ) )
            .forEach( ( st ) => this._listenToStore( publishers[ st ], st ) );
    }

    /**
     * Unregister from every store we listened to.
     */
    componentWillUnmount() {
        this._listener.stopListeningToAll();
    }

    /**
     * Starts listening to a store.
     * The state of the store will be connected to the state of the componentWillUnmount
     * @param  {Store} store
     * @param  {string} stateKey
     */
    _listenToStore( store: Store< * >, stateKey: string ) {
        this._listener.listenTo( store, ( value ) => this.setState( { [stateKey]: value } ) );
    }

    render() {
        const query = this.props.query || ( ( a ) => a );
        return this.props.render( query( this.state ) );
    }
}
